/** BlinkTask -- Example using schip scheduler
      with optional other core mod toys.

   With all the schip code installed this program:
    Turns the LED pin on and off for one second repeatedly.
    Reads ADC values and sends them to the serial port every 43ms.
    Accepts newline delimited text messages on the Serial port
     and echo's them back.

   The scheduler can be installed in the arduino cores directory
    (kernel version) or can be executed from the schip_scheduler
    library. But never shall the twain meet... In gerneral it's
    probably the library version so USEschipSCHEDlib should be
    defined below.

   The scheduler maintains a short list of functions to execute,
    each with a single int sized argument (which can be coerced to
    a 64Kb pointer if one likes), and the number of timer ticks
    (milli-sec) until it should be executed.
    
   The list is defined with a default length of eight items in:
     ...arduino/hardware/arduino/cores/[avr]/arduino/schip/scheduler.h
                or (more likely) in the library:
     Arduino/libraries/schip_scheduler/schip_scheduler.h

   The timer tick interrupt decrements the tick count on each pass.
    The standard Arduino loop() contains a call to the schip
    scheduler() function which checks for tick values of zero and
    then executes the relevant function with the given argument.
    To get the ball rolling, your setup() function needs to call:
     initScheduler()
   
   Tasks run in the "user space" and can do anything they want, as
    long as they don't block or delay(). Tasks can repost themselves
    or start other Tasks. Long running functions should be broken into
    smaller components and posted as a series of Tasks.

   Some additional ADC features are in library/schip_analog.
    The code installs an ADC interrupt handler and free runs each of the
    8 ADCs in sequence on a period of about 110 micro-sec. Channels 4,5
    are skipped because they are also used by the I2C interface, so only
    0-3,6,7 are converted. (Both the number of channels and the I2C skip
    thing can be controlled in the schip_analog.h header). The most
    recent channel result can be accessed immediately using a 'regular'
    analogRead() call. The first call to analogRead() gets the
    interrupts running and returns 0. This should be done in setup().

   Additionally, for each channel, the ADC code calculates an 8 bit
    average over 64 samples. At the end of each average cycle it
    posts the ADCTask() which the scheduler executes in "user space"
    (not as an interrupt). This task is posted about 23 times a second
    (or about 43 milli-sec per) on the standard 16MHz Arduino UNO and can
    be used to drive most human-interface-rate data collection schemes.
    The averaged ADC values are accessed with the function:
       analogReadAvg(i) where 'i' is the channel number 0-7
    with the proviso that channels 4 & 5 are skipped as above and will
    return a value of 0. The period of this average makes it useful for
    minimizing 50-60hz hum amongst other noises.
   
   Other ADC behaviors can be controlled with "SCHIP*" ifdefs in
    the header file. However the 64 sample period is a magic number
    and shouldn't be changed without fixing the magic elsewhere.
   
   If you have the "cores" HardwareSerial code installed:
   An interface has been added to the Serial class which allows
    the user to set a specific received message terminator character
    and then check if that terminator has been received. This allows
    messages to be received without the danger of blocking and waiting
    for the end of a message. The loop() code here calls the new
    msgAvailable() interface and posts messageTask() when a terminator
    is found. The messageTask() can do whatever it likes (with some care)
    in the way of reading and parsing the message.

   By necessity this code must be in the Arduino core, so many of the
    HardwareSerial files there are slightly modified. Without this feature
    users run the risk of blocking when receiving messages and thus making
    the Task scheduler run awkwardly. If you don't have it installed then
    SCHIPMSG will not be defined, so we can check this to use the old
    fashioned receive mechanism which just guesses when there's a message.
 
  This example code is in the public domain, FWIW.
**/

// get the prototypes for the schip scheduler additions
//  there is a secret redirect in this file that needs
//  to be set to use the "cores" version, otherwise it's
//   all in the schip_scheduler library
#include "schip_scheduler.h"

// comment this out to not use schip's analog imporvements
#include "schip_analog.h"
 
// Pin 13 has an LED connected on most Arduino boards.
#define LED 13

// Message termination character for text messages
#define MSGTERM '\n'

/** A simple task to be posted.
 *  This function matches the PFV typedef/signature for postTask()
 *   wiggles the LED pin and re-posts itself adinfinitum.
 *  The arg supplied in rePostTask() is sent to the next
 *   execution of myTask(arg) when the time comes.
 */
uint16_t delaytime = 1000;	// 'global' time between myTask() posts
void myTask( uint16_t arg )
{
    digitalWrite( LED, arg );   // wiggle the LED pin

#ifndef SCHIP_ANALOG_H	// NOT using schip_analog.h,
    					//  so ADCTask() isn't called automagically
    // call the ADC code ourselves just to see it work
    ADCTask( 8 );
#endif

    // repost this task,
    //  for 1000ms hence with inverted argument
    repostTask( delaytime, !arg );

    return;
}

/** utility to send a text formatted message without a newline appended
 *  @param d* -- integer data to send
 */
void sendTextn( int data1, int data2, int data3, int data4 )
{
    // print the values as ASCII with whitespace
    Serial.print( data1 );
    Serial.print( ' ' );
    Serial.print( data2 );
    Serial.print( ' ' );
    Serial.print( data3 );
    Serial.print( ' ' );
    Serial.print( data4 );
    Serial.print( ' ' );

    return;
}

/** utility to send a text formatted message WITH a newline appended
 *  @param d* -- integer data to send
 */
void sendText( int data1, int data2, int data3, int data4 )
{
    sendTextn( data1, data2, data3, data4 );
    Serial.println();

    return;
}

/** Task posted when ADC averaging gets a new set of values.
 **  Posted after every 64 conversions of all 6 channels
 **    @110us/channel ~= 43ms or 23 times a second
 **  Uses analogReadAvg( channel ); to get 8 bit results.
 ** @param numADC -- number of channels available -- required but ignored
 **
 **  This function reads both the averaged and some instantaneous results
 **   then sends them to the serial port as a newline delimited text message:
 **		avg0 avg1 avg2 avg3 avg4 avg5 avg6 avg7 ins0 ins1 ins2 ins3 
 **
 **  If not using schip_analog.h, then the old fashioned readAnalog()
 **   function will be used when this is called from the demo myTask()
 **/
void ADCTask( uint16_t numADC )
{
    int values[8];			// collected data values
    uint8_t i;				// utility counter

#ifdef SCHIPAVG // using schip's ADC library, with Averaging turned on
    // Note: if you have that library installed it will be posting ADCTasks
    //  so this gets executed no matter what. If the file is NOT installed
    //  then analogReadAvg() is not available and this function is only
    //  called from myTask()

    // ADC: read all 8 averaged results
    //  note that channels 4 & 5 are skipped due to I2C overlap
    for( i=0; i<numADC; ++i )
    {
        // read the averaged value for each input:
        values[i] = analogReadAvg( i );
    }

    // now print them
    sendTextn( values[0], values[1], values[2], values[3] );
    sendTextn( values[4], values[5], values[6], values[7] );

#endif // not USEschipADC

    //  Now the old fashioned way
    //  ADC: read first 4 instantaneous results
    for( i=0; i<4; ++i )
    {
        // read the instantaneous value for each input:
        values[i] = analogRead( i );
    }

    // and print them too, with a newline at the end
    sendText( values[0], values[1], values[2], values[3] );

    return;
}

/** Task posted when a serial message terminator is found
 *  @param nmsg -- number of message terminators in the input queue
 */
void messageTask( uint16_t nmsg )
{
    // EXAMPLE to see what's available:
    // int num = Serial.available();

/**
    // EXAMPLE to read the whole message string including '\n' terminator
    //  note that read... does NOT put a null string term in the buffer
    char buffer[21];
    int num = Serial.readBytesUntil( MSGTERM, buffer, 20 );
    buffer[num] = '\0';    // string terminator
    Serial.print( "got: <" );
    Serial.print( buffer );
    Serial.println( ">" );
**/

/**/
    // EXAMPLE to read one INT value and set "delaytime" LED flash period
    int val = Serial.parseInt();
    Serial.print( "got: " );
    Serial.println( val );
    delaytime = val;
    // consume everything else up to the terminator
    do
    {
        val = Serial.read();

    } while( (val != MSGTERM) && (val != -1) );
/**/

    return;
}

// the setup routine runs once when you press reset:
void setup()
{                
    // initialize the digital pin as an output.
    pinMode( LED, OUTPUT);    

      // start Serial at 9600baud
    Serial.begin( 9600 ); 

#ifdef SCHIPMSG 	// use schip's cores message receive code
    // set message terminator for text input
    Serial.setMsgTerm( MSGTERM );
#endif // SCHIPMSG

	// initialize schip's scheduler code (both library and cores version)
    initScheduler();

#ifdef SCHIP_ANALOG_H // using schip's ADC library
    // required to start the ADC interrupt cycle, returns 0 on first call...
    //  When running, the ADC averaging cycle will post ADCTask() every 43ms
    analogRead( 0 );
#endif // USEschipADC

    // post an initial task with an argument
    //  this will execute in 100ms: myTask(1)
    postTask( 100, myTask, 1 ); 
}

// the loop routine runs over and over again forever:
void loop()
{
    // USB message receive function
    //  if we have a new message, post the receive task
    //  if the message isn't handled by the time this runs again
    //   we will get a second Task post, so get with it...
	// if schip's message code is installed in kernel SCHIPMSG is defined
	//  so we can use it to get an full terminated message
	// otherwise we need to just check available()
	//   and hope we have the whole message
	word nMsg;
#ifdef SCHIPMSG 	// use schip's cores message receive code
    nMsg = Serial.msgAvailable();
#else				// use clunkier standard version
    nMsg = (Serial.available() > 1) ? 1 : 0; // more than one char available
#endif // not SCHIPMSG
	if( nMsg != 0 )
        postTask( 0, messageTask, nMsg );

    // check the scheduled task list and
    //  execute any whose time has come
    scheduler();
}
