/* testexec.ino -- Ardweenie main program, schip-MSCapture testing file.
		Delegates everything to the library files.
		Mostly so we can develop in a more reasonable environment.
		 Here we need to:
		  Include all the libraries we use;
		  Call the application setup routines in the library;
		  Then run the schip scheduler.
		   (supposedly everything is then a scheduled Task)
*/

/** Connect D9 or D11 to D8
 **
 **  This program starts a 500Hz PWM signal running on D11 at 50% duty cycle,
 **  and a slow 1ms/50ms signal running on D9.
 **  Then reads various periods from D8 using the MSCapture library.
 **
 **  The Serial port is initialized to 115200 baud.
 **  Input values are read and printed to Serial in the loop() function.
 **  three digit decimal numbers can be sent over the serial with a '\n'
 **   to set the pulse witdth of the D11 PWM signal
 **  
 **  With the base period for the D11 PWM signal being ~500Hz 
 **  I get a count of 32640 at full non-prescaled .0625uS speed, so ~490Hz.
 **  
 **  For no-prescale, MSCck1 ::
 **  With a base period for the Diddle output on D9 at ~20ms
 **  I get a count of ~327687 so 20.48ms (varying by about .5ms)
 **  
 **  With a base period for the Diddle output on D9 at ~40ms
 **  I get a count of ~655500 so 43.43ms (varying by about .5ms)
 **  With a 1ms ON period
 **  
 **  For /8 prescale, MSCck8 ::
 **  With a base period for the Diddle output on D9 at ~40ms
 **  I get a count of ~81900 so 40.95ms (varying by about .5ms)
 **  
 **  Testing Ultrasonic distance sensor:
 **  To calculate the range, where:
 **   D= Distance inch or centimeters
 **   Ew=Echo Width in uS (microseconds)
 **    Din=Ew/148
 **    Dcm=Ew/58 
 **  
 **  Range should vary around ~110->23200 uSec (2cm->400cm) with target,
 **  in 16Mhz counts that's 1760->371200.
 **  in 2Mhz counts that's 220->46400.
 **  
 **  If there is no target or out of range
 **  high output pulse Ew will terminate around 38mS
 **  
 **  But... with 50ms retriggers...???
 **   I seem to get 200ms (3,200,000 @16Mhz, 400,000 @2Mhz) failures,
 **  
 **/

// Arduino IDE needs to have this defined here...
//  include all libraries used by sub-contractors,
#include <Arduino.h>	// Our main stuff
#include <MSCapture.h>	// Our library files

// Get the prototypes for the schip-scheduler
//  (hidden in .../hardware/arduino/cores/arduino/schip
//   and accessed from default cores/arduino directory)
#include "schip/scheduler.h"

#define PWMPIN 		11	// PWM out run by analogWrite()
#define DIDDLEPIN 	 9	// PWM out run by scheduler timing
#define CAPPIN 	 8  // Capture in pin for ATmega 328p, cannot be changed
#define LED 	13	// LED output for trace

#define DIDDLEPERIOD	50	// diddleTask cycle in milli-sec
// #define AUTODODDLE	// define to repeat DIDDLEPIN cycle automatically

byte PWMval;	// current PWM output duty cycle value

// initialize the things we wish to fiddle upon
void runInit()
{
	Serial.println( "MSCapture Test Program" ); 
	Serial.println( "input:" ); 
	Serial.println( "value'\n' -- sets D11 PWM output ON length" );
}

// posted to read input when a message arrives
// then executes stuff based on input
void messageTask( unsigned int nmsg )
{
	// read INT values and get on with it
	//  expects to find some whitespace between multiple values and a
	//  newline at the end. If you accidentally put whitespace between the 
	//  last value and the newline this will timeout and hang everything
	//  the same thing happens if you hit return without any numbers...
	//  (this could be fixed with a peek() at the beginning of the loop)
	//  
	int vals[3]; // input values
	byte cnt = 0;
	int check;
	
	do
	{
		// get a value
		vals[cnt++] = Serial.parseInt();
/**
		Serial.print( "got: " );
		Serial.println( vals[cnt-1] );
**/
		// consume one char and see if it's whitespace
		check = Serial.read();

		// repeat with anything other than the terminating '\n'
	} while( (check != '\n') && (check != -1) && (cnt < 3) );
	
	// vals[] now contains cnt valid inputs
	// tell PWM to go to value 0
	PWMval = vals[0];
	analogWrite( PWMPIN, PWMval );

	return;
}

// scheduled task to turn on/off output pin with 1/DIDDLEPERIOD duty cycle
//  (for testing uSonic detector, this can trigger a ping
void diddleTask( word arg )
{
	// toggle the output pin
	digitalWrite( DIDDLEPIN, arg & 0x01 );

	// wiggle the LED so we know something is happening somewhere
	digitalWrite( LED, arg & 0x01 );

	if( arg & 0x01 )
	{
		// do 1ms ON pulse, increment arg
		repostTask( 1, ++arg );
	}

#ifdef AUTODODDLE
/** comment out to not repeat diddle....
**/
	else
	{
		// cycle every 50ms with a 1ms ON duty cycle
		repostTask( (DIDDLEPERIOD-1), ++arg );
	}
#endif AUTODODDLE

	return;
}
	
// the setup routine runs once when you press reset:
void setup()
{
	// everyone uses comms --
	//  use the fastest or else debug print hangs stuff up
	Serial.begin( 115200 ); 

	// set message terminator to newline for text input
    Serial.setMsgTerm('\n');

	// set pins to output
	pinMode(LED, OUTPUT);  
	pinMode(DIDDLEPIN, OUTPUT);  

	// start PWM and diddle output to test input counts
	pinMode(PWMPIN, OUTPUT);  	// probably redundant, but WTH...
	analogWrite( PWMPIN, 128);
	postTask( 0, diddleTask, 1 );

	// Initialize Capture [examples] -- prescale, mode
	//  where prescale is the micro-sec resolution starting 
	//   at 0.0625 (16MHz) to 64 (15.625 KHz) in steps of x8 
	//  and mode is which and how to measure the incoming pulses
	//   (see MSCapture.h for details
// 	MSCapture.begin( MSCck1, MSCmPERIOD );	// period btw rising edges
// 	MSCapture.begin( MSCck8, MSCmLPULSE );	// LOW pulse length
//	MSCapture.begin( MSCck64, MSCmAPULSE ); // HIGH/LOW duty cycle
// 	MSCapture.begin( MSCck256, MSCmHPULSE );	// HIGH pulse length
	MSCapture.begin( MSCck64, MSCmPERIOD ); // 4uS period measure

	// uncomment to start ADC interrupts
//	analogRead( 0 );

	// start schip task scheduler mechanism
	//  used by the PWM diddler
	initScheduler();

    runInit();		// init the system

    return;
}

// the loop() routine runs over and over and again forever:
void loop()
{
	byte tmp=0;

	// USB message receive function
	//  if we have a new message then post the receive task
	//  if the message isn't handled by the time this runs again
	//   we may get a second Task post, so get with it...
	byte nMsg = Serial.msgAvailable();
	if( nMsg != 0 )
		postTask( 0, messageTask, nMsg );

	// execute the schip task scheduler
	scheduler();	// schedule the world

/**/
	// check for capture result
	while( MSCapture.available() != 0 )
	{
		uint32_t val = MSCapture.read();
		if( val > 1000 )
			Serial.println("");

		Serial.print( val );
		Serial.println("");
//		Serial.print( " " );
		++tmp;
	}
/**/
		
	if( tmp )
	{
/*		to log PWM
		Serial.print( "PWM:" );
		Serial.println( PWMval );
*/
/*
		// or just terminate data line
		Serial.println("");
*/

#ifndef AUTODIDDLE
/** to only trigger when we got a signal from uSonic sensor
 **   also comment out else rePost() block in diddleTask()
**/
		postTask( 0, diddleTask, 1 );
#endif // not AUTODIDDLE
	}

	return;
}

// in case schipADC is defined
//  will be posted on ADC average cycle
void ADCTask( word numadc )
{
	int values[4];
	byte i;

	Serial.print( "ADC: " );

	// read four averaged ADC results
	//  will probably ignore channels 4 & 5 due to I2C use
	//  this is controlled by defines in cores/arduino/wiring_analog.c
	for( i=0; i<4; ++i )
	{
		// read the averaged value for each input:
		// values[i] = analogReadAvg( i );
		// read the instantaneous value for each input:
		values[i] = analogRead( i );
		Serial.print( values[i] );
		Serial.print( " " );
	}

	Serial.println( "" );

	return;
}

// in case SCHIPTCB is defined for the tick interrupt
void schipTick(void)
{
	return;
}
