/* example.ino -- Arduino main program to demo schip-improvements.
**		Delegates almost everything to the schip library files
**       schip_scheduler.h and schip_analog.h
**		Here we need to:
**		  Include all the libraries we use;
**		  Call the application setup routines in the library;
**		  Post some tasks to do some stuff sometimes.
**		  Then run the schip scheduler.
**		   (supposedly everything is then a scheduled Task)
**
**	This program does two (or three) things:
**   1. The scheduler is used to run four instances of UpdateTaskN() 
**    each with an argument that specifies the state of an output pin.
**	  The Task toggles the secified pin on/off at a rate that
**    depends on the pin number. The rates are in the audio range
**    so listenting to pins 9-12 should hear different pitches.
**	 2. The ADC interrupt and averaging code posts the ADCTask()
**	  approximately every 43ms, when 6 of the possible 8 ADCs are
**    being converted. The task reads the first four averaged
**	  ADC values and sends them to the serial port at 9600baud.
**	 3. (If SCHIPBUFFER is defined in schip/schip_analog.h) The loop
**	  function checks SchipBuf0 for a full buffer of ADC0 data and
**    loops through it converting it to 16 bit ints. SchipBuf0 gets
**	  set by the ADC interrupt code on the same period as ADCTask().
**    See README.txt in schip_analog library for details.
*/

// Arduino IDE needs to have this defined here...
//  include all libraries used by sub-contractors,
#include <Arduino.h>	  // everybody wants this

// get the schip ADC additions
//  in the schip_analog library
#include "schip_analog.h"

// Get the prototypes for the schip scheduler additions
#include "schip_scheduler.h"

// digital I/O ports we will use, just for Tasking demos
#define OUT0	 8
#define OUT1	 9
#define OUT2	10
#define OUT3	11
#define LED		13

// note: logging may mess with scheduling accuracy...YMMV
#define LOGADC		// define to send ADC results to serial port

// note: this is NOT the number that are being converted, see NUM_ADCCHANNEL
#define NUMADC	 4	// number of ADC channels to log

/** Task functions to be used in demo **/

/** Task posted when ADC averaging gets a new set of values.
 **  Posted from ADC code after every 64 conversions of all channels
 **   @110us/channel ~= 43ms or 23 times a second when 6 channels are used
 **  Use analogReadAvg( channel ); to get 8 bit averaged results.
 ** @param numADC -- number of channels available -- required but ignored here
 **/
void ADCTask( uint16_t numADC )
{
	static byte sCnt = 0;	// global utility variable
	int value;				// ADC value
	byte i;					// convenient index

	++sCnt; // global running count of sample cycles we have seen

	// alternate LED on/off so we can see this working
	digitalWrite( LED, sCnt&0x01 );

	// log the current "time"
	unsigned long mytime = millis();	// start time...
	Serial.print( mytime );
	Serial.print( " -- ADC: " );

#ifdef LOGADC	// if defined, print values

#ifdef SCHIPAVG	// using the schip_analog averaging code
	// read first four averaged ADC results
	for( i=0; i<NUMADC; ++i )
	{
		// read and log the averaged value of input:
		value = analogReadAvg( i );
		Serial.print( value );
		Serial.print( " " );
	}

	Serial.print( "-- " );
#endif // SCHIPAVG

	// Then read first four immediate ADC results
	for( i=0; i<NUMADC; ++i )
	{
		// read and log the immediate value of input:
		value = analogRead( i );
		Serial.print( value );
		Serial.print( " " );
	}

#endif // LOGADC

	Serial.println();

#ifndef SCHIPAVG	// NOT using the schip Averaging code in schip_analog
	// repost ADCTask() to just repeat itself periodically (arg==0)
	repostTask( 1000, 0 );	// run again in 1 sec
#endif

	return;
}

/** Just some sample Tasks to do stuff...
** These wiggle output pins on different cycles.
**   repost self with different execution delays
**  @param arg -- new pin state, 0 or 1
**/
// do the real work, set pin to state and repost in ticks
void commonTask( byte pin, byte state, byte ticks )
{
	// wiggle our output pin
	digitalWrite( pin, state );

	// repost to execute in ticks milli-sec with inverted argument
	repostTask( ticks, !state );

	return;
}

// Task for OUT0
void UpdateTask0( word arg )
{
	// wiggle the output pin and repost in N ticks
	commonTask( OUT0, arg, 2 );
	return;
}

// Task for OUT1
void UpdateTask1( word arg )
{
	// wiggle the output pin and repost in N ticks
	commonTask( OUT1, arg, 3 );
	return;
}

// Task for OUT2
void UpdateTask2( word arg )
{
	// wiggle the output pin and repost in N ticks
	commonTask( OUT2, arg, 4 );
	return;
}

// Task for OUT3
void UpdateTask3( word arg )
{
	// wiggle the output pin and repost in N ticks
	commonTask( OUT3, arg, 5 );
	return;
}


// the setup routine runs on power up or when you press reset:
void setup()
{
	initScheduler();		// initialize Task scheduling
	Serial.begin(9600);		// serial port logging at 9600baud

	// initialize the led and tracking pins as outputs
	pinMode(LED, OUTPUT);  
	pinMode(OUT0, OUTPUT);  
	pinMode(OUT1, OUTPUT);  
	pinMode(OUT2, OUTPUT);  
	pinMode(OUT3, OUTPUT);  

	// start tasks to wiggle each output pin on different cycles
	//  postTask( ticks-to-execute, function, argument )
	postTask( 0, UpdateTask0, 0 );
	postTask( 0, UpdateTask1, 0 );
	postTask( 0, UpdateTask2, 0 );
	postTask( 0, UpdateTask3, 0 );

#ifndef SCHIPAVG	// NOT using the schip Averaging code in schip_analog
	// post the ADCTask() to just repeat itself periodically (arg==0)
	postTask( 100, ADCTask, 0 );
#endif

	// required to start the ADC interrupt cycle
	// which will eventually repeatedly post ADCTask()
	//  and/or toggle SchipBuf0 semaphore
	analogRead( 0 );

    return;
}

// the loop() routine runs over and over and again forever:
void loop()
{
	// execute the schip task scheduler
	//  to schedule the world
	scheduler();

#ifdef SCHIPBUFFER // if Buffering is defined in schip_analog.h
	// example using the ADC0 input buffering feature
	// note that doing this WILL mess with scheduler timing as well
	// ADC code will set SchipBuf0 to point to data buffer when ready
	if( SchipBUF0 != NULL )
    {
		// for instance....
		// rummage through the buffer and convert to 16 bit signed data
		for( int i = 0 ; i < SBUFSIZE ; ++i ) // copy samples
		{
			int k = SchipBUF0[i]; // get 10 bit unsigned ADC data
			k -= 0x0200;	// make a signed int around nominal zero point
			k <<= 6;		// promote to 16b signed int
		}

		// reset semaphore, optional but good practice
		SchipBUF0 = NULL;

		// log the current "time"
		unsigned long mytime = millis();
		Serial.print( mytime );
		Serial.println( " -- Buffer done" );
	}
#endif // SCHIPBUFFER example

	return;
}
