/*
  TuneTest.ino -- Arduino main program for Schip's Tunes demo.
   Plays a tune using tone() function on pin 11 using scheduled Tasks
   Also logs events to Serial at 9600 baud in text or MIDI events.
*/

// Get system stuff
#include <Arduino.h>

// Get the prototypes for the schip-scheduler
//  note: can switch to "cores" scheduler version in that file
#include <schip_scheduler.h>

// include some local stuff
#include "./Tunes.h"

// define this to output MIDI notes on Serial rather than tone() & Text log
//  note that the "pitch" values in the tune files will need to be
//   converted to MIDI note numbers for this to work "right"
//#define MIDIOUTPUT

// pins to use
#define LED		   13	// onboard LED, just for convenience
#define SPEAKERout 11	// the pin to which the speaker is attached

#define useADC	    0	// use ADC 0 for analog input
#define useIN0	    5	// use pin 5 for digital input
#define useIN1	    6	// use pin 6 for digital input
// comment out SAMPLETICKS to not do inputTask()
#define SAMPLETICKS	100	// run inputTaks() every NNNms


Note *Melody = 0;	// array of Notes we are playing


/* Play a tone at pitch for dticks milli-seconds
**   and optionally log Text or MIDI ON Serial event
**/
void noteOn( uint16_t ind, uint16_t pitch, uint16_t dticks, uint16_t lticks )
{
#ifdef MIDIOUTPUT // do MIDI instead of tone()
    // send a real MIDI note ON message
    Serial.write(144);
    Serial.write(pitch);
    Serial.write(127);
#else
    // play the tune on the output pin
    //  if you can figure out how to convert pitch to MIDI note number
    //   you could do both this and the MIDI out above. be smart...
    tone( SPEAKERout, pitch, dticks );

    // log a pseudo MIDI note ON message
    Serial.print( ind );
    Serial.print( " [" );
    Serial.print( millis() );
    Serial.print( "] " );
    Serial.print( pitch );
    Serial.print( " On (" );
    Serial.print( dticks );
    Serial.print( " " );
    Serial.print( lticks );
    Serial.println( ")" );
#endif
}

/* log Text or MIDI OFF Serial event
**/
void noteOff( uint16_t pitch )
{
#ifdef MIDIOUTPUT // do MIDI instead of tone()
    // send a real MIDI note OFF message
    Serial.write(128);
    Serial.write(pitch);
    Serial.write(127);
#else
    //log a pseudo MIDI note OFF message
    Serial.print( "  [" );
    Serial.print( millis() );
    Serial.print( "] " );
    Serial.print( pitch );
    Serial.println( " Off" );
#endif
}

/** notePlay - Starts playing with NoteOn and posts a noteOff event.
 **
 ** @param ind -- index into sequence, just for logging
 ** @param pitch -- frequency of note to play
 ** @param dticks -- duration of note on in 1ms ticks
 **/
void notePlay( uint16_t ind, uint16_t pitch, uint16_t dticks, uint16_t lticks )
{
    // send note on message
    noteOn( ind, pitch, dticks, lticks );

    // Schip's scheduler replaces delay
    // note: postTask( uint16_t ticks, PFV func, uint16_t arg );
    postTask( dticks, noteOff, pitch );

    return;
}

// Task posted for each Melody note,
// @param ind -- index into Melody sequence
static void TuneTask( uint16_t ind )
{
	// if nothing to do, don't do it
	//  if Melody is not set (==0) there's nothing to play
	//  note: can stop the playing by setting Melody = 0
	if( Melody == 0 ) 
	{
		return;
	}

	// toggle LED so maybe we see this working
	digitalWrite( LED, !(ind&0x01) );

	// one Melody Note contains:
	//  pitch -- buzzer frequency (or optionally MIDI note number)
	//  duration -- ON time of this note 
	//  length -- time until next event
	//   duration and length are in 8ms increments
	//    so we have to multiply by 8 (shift up, << 3) to get tick counts

	// turn on the current note if it's not 0
	if( Melody[ind].pitch != REST )
		notePlay( ind, Melody[ind].pitch,
				((uint16_t)Melody[ind].duration) << 3,
				((uint16_t)Melody[ind].length << 3) );

	if( Melody[ind].length == DONE )
	{
		// clear Melody so we don't accidentally play something
		Melody = 0;
		// wiggle LED off just in case
		digitalWrite( LED, 0 );
		Serial.println( "done" );
	}
	else if( Melody[ind].length == REPEAT )
	{
		// rewind
		// repost this task in one second
		//  with "ind" set to 0 again
		repostTask( 0, 0 );
		Serial.println( "repeat" );
	}
	else
	{
		// play the next note next time
		// repost this task in length ticks
		//  with "ind" incremented to next note in sequence
		// note: MUST use an intermediate for either next ticks
		//   or incremented ind because the incr gets evaluated
		//   before the array index that uses it if you put them
		//   both in the repostTask() line itself... ugh...
		uint16_t next = ((uint16_t)Melody[ind].length << 3);
		repostTask( next, ++ind );
	}

	return;
}

// start playing immediately at Note number 0 of specified m Melody
void startTune( Note *m )
{
	Melody = m;
	postTask( 0, TuneTask, 0 );
	return;
}

/** Example Task to sample some inputs on a regular basis.
 **  arg is the sample period in ms ticks to use
 **  sets global values on each pass:
 **/
int adc0 = 0;
int in0  = 0;
int in1  = 0;
void inputTask( uint16_t arg )
{
	// read some inputs
	in0 = digitalRead( useIN0 ); // nominally: pin 5
	in1 = digitalRead( useIN1 ); // pin 5
	adc0 = analogRead( useADC ); // adc 0

	Serial.print( "input: " );
	Serial.print( adc0 );
	Serial.print( " " );
	Serial.print( in0 );
	Serial.print( " " );
	Serial.println( in1 );
	

	// re-execute in "arg" milli-sec with "arg" as the argument
	repostTask( arg, arg );

	return;
}

/** No schip ADCs were used in the making of this demo program
 **  but we may need to provide a stub for schip's ADC interrupt Task
 **  because the Argweenie compile system is too dumb to not include it
 **/
void ADCTask( uint16_t arg )
{
	Serial.println( "ADCTask()" );
	return;
}


// the setup routine runs once when you press reset:
void setup()
{  
	initScheduler();
	Serial.begin(9600);	// just in case we want to print something

	// set our I/O pins
	pinMode( useIN0, INPUT_PULLUP );
	pinMode( useIN1, INPUT_PULLUP );
	pinMode( LED, OUTPUT );
	pinMode( SPEAKERout, OUTPUT );
	
	// initialize schip's scheduler
	initScheduler();

#ifdef SAMPLETICKS
	// start the input sampling task
	postTask( 0, inputTask, SAMPLETICKS );
#endif

	// start playing a tune
//	startTune( Lullaby );
//	startTune( HatDance );
	startTune( Twinkle );

	return;
}

// the loop() routine runs over and over and again forever:
void loop()
{
	// execute the schip task scheduler
	scheduler();	// schedule the world
	return;
}
