/* stepint.ino -- Stepper motor driver driven from tick interrupt.
**  Hardcodes stepper motor attached to pins 2-5 of UNO type board.
**
**  Uses schip's kernel core scheduler() with SCHIPTCB defined
**  Accepts commands from Serial at 9600 baud,
**   with a NEWLINE terminator.
**/

// Get all the Arduino system definitions
//  by including this file from the development system directories
#include <Arduino.h>
// may also need these internal header files for defines
//#include "wiring_private.h"
//#include "pins_arduino.h"

// get the prototypes for the schip scheduler additions
//  (hidden in arduino.../hardware/arduino/cores/arduino/schip
//   and accessed from default cores/arduino directory...
//   OR in libraries/schip_scheduler)
// note: This code requires the cores version for schipTick() 
//#include "schip_scheduler.h" // DO NOT USE schip's scheduler as a library
#include <schip/scheduler.h> // use schip's scheduler installed in cores

// select one step operation mode
//#define FULLSTEP
#define HALFSTEP

#define STEPTRACE // use LED to show when motor is running

#define PROFILETICKS 15 // number of timer ticks to do startup profile


// Message termination character for text commands
#define MSGTERM '\n'

// stepper motor control pins -- on UNO Port D
#define STEP1PIN	2	// Port D, Bit 2 - mask 0x04
#define STEP1MASK	0x04
#define STEP2PIN	3	// Port D, Bit 3 - mask 0x08
#define STEP2MASK	0x08
#define STEP3PIN	4	// Port D, Bit 4 - mask 0x10
#define STEP3MASK	0x10
#define STEP4PIN	5	// Port D, Bit 5 - mask 0x20
#define STEP4MASK	0x20

// all the stepper bits
#define STEPBITMASK	(STEP1MASK | STEP2MASK | STEP3MASK | STEP4MASK)

#define LED			13


#if 0
/** Posted from ADC interrupt code,
 **  gets and prints all the relevant input values
 **/
void ADCTask( uint16_t numADC )
{
	int values[12];	// collected data values
	uint8_t i;		// utility counter

/**/
	// ADC: read first 4 averaged results
	for( i=0; i<4; ++i )
	{
		// read the averaged value for each input:
		values[i] = analogReadAvgMS( i );
	}
/**/
/**/
	// ADC: read first 4 non-averaged results
	for( i=4; i<8; ++i )
	{
		// read the instantaneous value for each input:
		values[i] = analogReadMS( i-4 );
	}
/**/

	// read the digital inputs into values[7]
	i = 0;
	i |= digitalRead( dLS2 );	// dLS* end up in top nibble
	i <<= 1;
	i |= digitalRead( dLS1 );
	i <<= 2;
	i |= digitalRead( D4 );		// unused inputs in bottom nibble
	i <<= 1;
	i |= digitalRead( D3 );
	i <<= 1;
	i |= digitalRead( D2 );
	values[7] = i;

	// now test the input class methods
	values[8] = roverinput.lineSensors();
	values[9] = roverinput.encoders();
	values[10] = roverinput.light();
	values[11] = roverinput.distance();

	if( Printvals )
	{
		// print it all, first AVG ADC values, w/o newline
		plot.textn( values[0], values[1], values[2], values[3] );
		// rover input class massaged values, w/o newline
		plot.textn( values[8], values[9], values[10], values[11] );
		// raw ADC set with last one being hex digital INs, with newline
		plot.textX( values[4], values[5], values[6], values[7] );
	}

	return;
}
#endif


// global control values:
byte step2do = 0;	// steps to execute -- 0 = none, 255 = forever
byte tick2do = 0;	// final number of timer ticks between steps
#if 0
char nticks = 0;	// current number of timer ticks between steps
#endif
byte nticks = 0;	// current number of timer ticks between steps
byte tickcnt = 0;	// number of timer ticks executed -- decr to 0
byte stepstate = 0;	// state bits, see below for values
#define sDIR  0x01	// bottom bit selects direction
#define sHLD  0x02	// if set, then leave step pins ON at end, else all OFF
#define sPRF  0x04	// if set, then profile motor startup
#define sRUN  0x80	// motor running when set

char stepnum = 0;	// motor step number in sequence -- 0-3 or 0-7

/** Start the stepper control
 **  @param dir   -- direction 1,0 arbitrary YMMV
 **  @param steps -- number of steps to do 0=stop, 255=forever
 **  @param ticks -- 1ms timer ticks between steps
 **  @param prf   -- if set, profile motor startup (for low tick counts only)
 **  @param hold  -- 0= shutoff coils, else hold state of all coil-pins at end
 **
 **   note: the step count does not account for FULL/HALF setting
 **			so should be doubled if using HALF-steps
 **/
void startStepper( byte dir, byte steps, byte ticks, byte prf, byte hold )
{
	byte xnstate = ( sRUN |
				  (dir  ? sDIR : 0) |
				  (prf  ? sPRF : 0) |  // not actually used, but...
				  (hold ? sHLD : 0) );
	ticks -= 1; // note: account for post decrement when running...
	//todo: doesn't deal right with high tick2do counts (vis wraparound)...
	byte xticks = ( prf ? (ticks + PROFILETICKS+1) : ticks );
	
	// todo: probably need interrupt disable around all this...
	// shut off interrupts to set values used by tick interrupt
	byte oldSREG = SREG;	// save interrupt settings
	cli();
	stepstate = xnstate;
	step2do = steps;
	tick2do = ticks;
	nticks = xticks;
	tickcnt = 0;	// forces first step on first call below
	SREG = oldSREG; // restore interrupt settings

	return;
}

/** Returns the number of steps left to do, or 0 when done
 **/
byte getSteps()
{
	// single byte access, so no interrupt mangling needed
	return step2do;
}

/** stop operation immediately
 **  @param hold  -- 0= shutoff coils, else hold state of all coil-pins at end
 **/
void stopStepper( byte hold )
{
	// todo: definately need interrupt disable around this...
	byte oldSREG = SREG;	// save interrupt settings
	cli();
	stepstate = 0;
	step2do = 0;
	SREG = oldSREG; // restore settings

	if( hold == 0 )
		PORTD &= ~STEPBITMASK;	// clear coils

	return;
}

/** Function called from kernel tick interrupt in wiring.c w/ SCHIPTCB set
 **  manages counters and stepper motor steps.
 **/
void schipTick()
{
	if( !(stepstate) )	// note, at minimum sRUN will be set when running
	{
		// if not running, don't do anything
		return;
	}
	else if( (tickcnt != 0) )
	{
		// not ready for a step yet
		--tickcnt;
		return;
	}

	// if this is the last step in the run,
	//  just cleanup and reset
	if( !(step2do) )
	{
		// we waited for the last tickcnt to expire,
		//  shutoff step output bits if not HOLD
		//  and clear state so we don't do it again...
		if( !(stepstate & sHLD) )
		{
			//  shut off coils,
			//  otherwise we will leave coils charged
			PORTD &= ~STEPBITMASK;
		}

		stepstate = 0;

#ifdef STEPTRACE
		// turn LED OFF on UNO
		PORTB &= ~(1<<PINB5);
#endif // STEPTRACE

		return;
	}

	// ...OK now we can do a real step...

#ifdef STEPTRACE
	// toggle LED ON on UNO
	PORTB ^= (1<<PINB5);
#endif // STEPTRACE

	// shut all bits OFF (briefly)
	PORTD &= ~STEPBITMASK;

#ifdef FULLSTEP
	// set the appropriate bits ON for full single-coil stepping
	switch( stepnum )
	{
	  case 0:
		PORTD |= STEP1MASK;
	  break;

	  case 1:
		PORTD |= STEP2MASK;
	  break;

	  case 2:
		PORTD |= STEP3MASK;
	  break;

	  case 3:
		PORTD |= STEP4MASK;
	  break;

	  default:
	  	stepnum = 0;
	  break;
	}

	// recycle at end
	if( stepstate & sDIR )
	{
		if( ++stepnum >= 4 )		stepnum = 0;
	}
	else
	{
		if( --stepnum < 0 )		stepnum = 3;
	}
		
#endif // FULLSTEP

#ifdef HALFSTEP
	// set the appropriate bits ON for half stepping
	switch( stepnum )
	{
	  case 0:
		PORTD |= STEP1MASK;
	  break;

	  case 1:
		PORTD |= (STEP1MASK | STEP2MASK);
	  break;

	  case 2:
		PORTD |= STEP2MASK;
	  break;

	  case 3:
		PORTD |= (STEP2MASK | STEP3MASK);
	  break;

	  case 4:
		PORTD |= STEP3MASK;
	  break;

	  case 5:
		PORTD |= (STEP3MASK | STEP4MASK);
	  break;

	  case 6:
		PORTD |= STEP4MASK;
	  break;

	  case 7:
		PORTD |= (STEP4MASK | STEP1MASK);
	  break;

	  default:
	  	stepnum = 0;
	  break;
	}

	// increment step to do and recycle at end
	if( stepstate & sDIR )
	{
		if( ++stepnum >= 8 )
		{
			stepnum = 0;

#if 0 // old profile method
			// deal with profiling if so
			nticks -= 2;
			if( nticks < tick2do )
				nticks = tick2do;
#endif
		}
	}
	else
	{
		if( --stepnum < 0 )
		{
			stepnum = 7;

#if 0 // old profile method
			// deal with profiling if so
			nticks -= 2;
			if( nticks < tick2do )
				nticks = tick2do;
#endif
		}
	}

#endif // HALFSTEP

	// profile: ramp up motor 'speed' on each step during startup 
	if( nticks > tick2do )	--nticks;

	tickcnt = nticks;		// restart tick counter
	//tickcnt = tick2do;	// restart tick counter, non-profiled

	if( step2do != 255 )	// keep track of steps done
		--step2do;

	return;
}


/** Task posted when a serial message terminator is found
 *  @param nmsg -- number of message terminators in the input queue
 */
void messageTask(uint16_t nmsg)
{
	static byte hold = 0;	// hold step state when run finishes
	static byte prf = 0;	// do motor startup profiling
	int steps;	// input value for number of steps to execute
	int speed;	// input value for millisec speed setting
	int pin;	// input value
	char val = Serial.read(); // get the command character
	
	Serial.print( "cmd "  );	// see what we got, for debugging, etc

	// Perform an action depending on the command
	switch (val)
	{
	  case 'r': // run forward a bit just for easy test
	  case 'R':
		Serial.println( "Run: fwd 100 steps at 10ms" );
		// startStepper( byte dir, byte steps, byte ticks, byte prf, byte hold )
		startStepper( 1, 100, 10, prf, hold );
	  break;

	  case 'f': //Move Forward
	  case 'F':
		steps = (byte) Serial.parseInt();
		speed = (byte) Serial.parseInt();
		Serial.print( "Fwd: " );
		Serial.print( steps );
		Serial.print( " steps at " );
		Serial.println( speed );
		startStepper( 1, steps, speed, prf, hold );
	  break;

	  case 'b': //Move Backwards
	  case 'B':
		steps = (byte) Serial.parseInt();
		speed = (byte) Serial.parseInt();
		Serial.print( "Bkd: " );
		Serial.print( steps );
		Serial.print( " steps at " );
		Serial.println( speed );
		startStepper( 0, steps, speed, prf, hold );
	  break;

	  case 'h': // toggle HOLD on and off for next command
	  case 'H':
		hold ^= 0x01;
		Serial.print( "Hold: " );
		Serial.println( hold ? "ON" : "OFF" );
	  break;

	  case 'p': // toggle Profile on and off for next command
	  case 'P':
		prf ^= 0x01;
		Serial.print( "Profile: " );
		Serial.println( prf ? "ON" : "OFF" );
	  break;

	  case 'g': //Get remaining stepcount
	  case 'G':
		steps = getSteps();
		Serial.print( "Steps: " );
		Serial.println( steps );
	  break;

	  case 'd': // Set specified pin ON, use Stop to shut off
	  case 'D':
		pin = (byte) Serial.parseInt();
		Serial.print( "Dig ON: " );
		Serial.println( pin );
		digitalWrite( pin, 1 );
	  break;

	  case 's': //Stop
	  case 'S':
	  default: // also stop for any other character
		Serial.println( "Stop" );
		stopStepper( hold );
	  break;
	}
	
	 // consume everything else up to the terminator
	do
	{
		val = Serial.read();

	} while ((val != MSGTERM) && (val != -1));

	return;
}

// The Arduino system requires these two functions, setup() and loop()

//  setup() runs once after a Start or Reset
void setup()
{
	// set each pin's control direction
	//  done in constuctors for motion and input classes

#ifdef SCHIPMSG // use schip's kernel message receive task code
	// set message terminator for text input
	Serial.setMsgTerm( MSGTERM );
#endif

	// set pins
	pinMode( LED, OUTPUT );
	pinMode( STEP1PIN, OUTPUT );
	pinMode( STEP2PIN, OUTPUT );
	pinMode( STEP3PIN, OUTPUT );
	pinMode( STEP4PIN, OUTPUT );

	initScheduler();

	// set the serial communication port to 9600 baud
	//  note: use 9600 value and newline terminator in the IDE as well
	Serial.begin( 9600 );
	
	// required to start ADC interrupt code running
	//  data collection is driven by repeated posts of ADCTask()
	//   made by the ADC interrupt code
	//analogRead(0);
	
	// send some explanation to the PC host
	Serial.println( "Setup Done, {f,b} steps ticks" );

	return;
}

//  loop() runs repeatedly until the system is turned off
void loop()
{
    // USB message receive function
    //  if we have a new message, post the receive task
    //  if the message isn't handled by the time this runs again
    //   we will get a second Task post, so get with it...
	// if schip's message code is installed in kernel
	//  we can use it to get an full terminated message
	// otherwise we need to just check available()
	//   and hope we have the whole message

	word nMsg;
#ifdef SCHIPMSG // use schip's kernel message receive task code
    nMsg = Serial.msgAvailable();
#else					// use clunkier standard version
    nMsg = (Serial.available() > 1) ? 1 : 0; // more than one char available
#endif // not SCHIPMSG
	if( nMsg != 0 )
        postTask( 0, messageTask, nMsg );

    // check the scheduled task list and
    //  execute any whose time has come
    scheduler();
}
