/* steptest.ino -- Stepper motor driver test program
**  uses stepint library driven from tick interrupt.
**  Hardcodes stepper motor attached to pins 2-5 of UNO type board.
**
**  Uses schip's kernel core scheduler() with SCHIPTCB defined
**  Accepts commands from Serial at 9600 baud,
**   with a NEWLINE terminator.
**/

// Get all the Arduino system definitions
//  by including this file from the development system directories
#include <Arduino.h>

// get the prototypes for the schip scheduler additions
//  (hidden in arduino.../hardware/arduino/cores/arduino/schip
//   and accessed from default cores/arduino directory...
//   OR in libraries/schip_scheduler)
// note: This code requires the cores version for schipTick() 
//#include "schip_scheduler.h" // DO NOT USE schip's scheduler as a library
#include <schip/scheduler.h> // use schip's scheduler installed in cores

#include "stepint.h"	 // schip's stepper library


// Message termination character for text commands
#define MSGTERM '\n'

#define LED			13

/** Task posted when a serial message terminator is found
 *  @param nmsg -- number of message terminators in the input queue
 */
void messageTask(uint16_t nmsg)
{
	static byte hold = 0;	// hold step state when run finishes
	static byte prf = 0;	// do motor startup profiling
	int steps;	// input value for number of steps to execute
	int speed;	// input value for millisec speed setting
	int pin;	// input value
	char val = Serial.read(); // get the command character
	
	Serial.print( "cmd "  );	// see what we got, for debugging, etc

	// Perform an action depending on the command
	switch (val)
	{
	  case 'r': // run forward a bit just for easy test
	  case 'R':
		Serial.println( "Run: fwd 100 steps at 10ms" );
		// startStepper( byte dir, byte steps, byte ticks, byte prf, byte hold )
		startStepper( 1, 100, 10, prf, hold );
	  break;

	  case 'f': //Move Forward
	  case 'F':
		steps = (byte) Serial.parseInt();
		speed = (byte) Serial.parseInt();
		Serial.print( "Fwd: " );
		Serial.print( steps );
		Serial.print( " steps at " );
		Serial.println( speed );
		startStepper( 1, steps, speed, prf, hold );
	  break;

	  case 'b': //Move Backwards
	  case 'B':
		steps = (byte) Serial.parseInt();
		speed = (byte) Serial.parseInt();
		Serial.print( "Bkd: " );
		Serial.print( steps );
		Serial.print( " steps at " );
		Serial.println( speed );
		startStepper( 0, steps, speed, prf, hold );
	  break;

	  case 'h': // toggle HOLD on and off for next command
	  case 'H':
		hold ^= 0x01;
		Serial.print( "Hold: " );
		Serial.println( hold ? "ON" : "OFF" );
	  break;

	  case 'p': // toggle Profile on and off for next command
	  case 'P':
		prf ^= 0x01;
		Serial.print( "Profile: " );
		Serial.println( prf ? "ON" : "OFF" );
	  break;

	  case 'g': //Get stepcounts
	  case 'G':
		steps = getSteps2do();
		Serial.print( "Steps2do: " );
		Serial.print( steps );
		steps = getStepsDone();
		Serial.print( "  Steps done: " );
		Serial.println( steps );
	  break;

	  case 'd': // Set specified pin ON, use Stop to shut off
	  case 'D':
		pin = (byte) Serial.parseInt();
		Serial.print( "Dig ON: " );
		Serial.println( pin );
		digitalWrite( pin, 1 );
	  break;

	  case 's': //Stop
	  case 'S':
	  default: // also stop for any other character
		Serial.println( "Stop" );
		stopStepper( hold );
	  break;
	}
	
	 // consume everything else up to the terminator
	do
	{
		val = Serial.read();

	} while ((val != MSGTERM) && (val != -1));

	return;
}

// The Arduino system requires these two functions, setup() and loop()

//  setup() runs once after a Start or Reset
void setup()
{
	// set each pin's control direction
	//  done in constuctors for motion and input classes

#ifdef SCHIPMSG // use schip's kernel message receive task code
	// set message terminator for text input
	Serial.setMsgTerm( MSGTERM );
#endif

	// set pins
	pinMode( LED, OUTPUT );
	pinMode( STEP1PIN, OUTPUT );
	pinMode( STEP2PIN, OUTPUT );
	pinMode( STEP3PIN, OUTPUT );
	pinMode( STEP4PIN, OUTPUT );

	initScheduler();

	initStepper();

	// set the serial communication port to 9600 baud
	//  note: use 9600 value and newline terminator in the IDE as well
	Serial.begin( 9600 );
	
	// required to start ADC interrupt code running
	//  data collection is driven by repeated posts of ADCTask()
	//   made by the ADC interrupt code
	//analogRead(0);
	
	// send some explanation to the PC host
	Serial.println( "Setup Done, {f,b} steps ticks" );

	return;
}

//  loop() runs repeatedly until the system is turned off
void loop()
{
    // USB message receive function
    //  if we have a new message, post the receive task
    //  if the message isn't handled by the time this runs again
    //   we will get a second Task post, so get with it...
	// if schip's message code is installed in kernel
	//  we can use it to get an full terminated message
	// otherwise we need to just check available()
	//   and hope we have the whole message

	word nMsg;
#ifdef SCHIPMSG // use schip's kernel message receive task code
    nMsg = Serial.msgAvailable();
#else					// use clunkier standard version
    nMsg = (Serial.available() > 1) ? 1 : 0; // more than one char available
#endif // not SCHIPMSG
	if( nMsg != 0 )
        postTask( 0, messageTask, nMsg );

    // check the scheduled task list and
    //  execute any whose time has come
    scheduler();
}
